<?php

namespace Drupal\wpf;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Image\ImageFactory as CoreImageFactory;
use Drupal\Core\Image\ImageInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\file\FileInterface;
use Drupal\system\Plugin\ImageToolkit\GDToolkit;

/**
 * ImageFactory service for creating fallback image.
 */
class ImageFactory implements ImageFactoryInterface {

  use StringTranslationTrait;

  /**
   * The image factory service.
   *
   * @var \Drupal\Core\Image\ImageFactory
   */
  protected $imageFactory;

  /**
   * The logger channel factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $logger;

  /**
   * The ConfigFactoryInterface definition.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The EntityTypeManagerInterface definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The FileSystemInterface definition.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The image quality factor.
   *
   * @var int
   */
  protected $quality;

  /**
   * The original image uri.
   *
   * @var string
   */
  protected $uri;

  /**
   * The fallback image uri.
   *
   * @var string
   */
  protected $destinationUri;

  /**
   * The image toolkit.
   *
   * @var \Drupal\Core\ImageToolkit\ImageToolkitInterface
   */
  protected $toolkit;

  /**
   * The toolkit id definition.
   *
   * @var string
   */
  protected $toolkitId;

  /**
   * The source image.
   *
   * @var \Drupal\Core\Image\ImageInterface
   */
  protected $sourceImage;

  /**
   * The fallback image mime type.
   *
   * @var string
   */
  protected $mimeType;

  /**
   * Constructs an ImageFactory object.
   *
   * @param \Drupal\Core\Image\ImageFactory $image_factory
   *   The image.factory service.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger
   *   The logger channel factory.
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translation
   *   The string translation service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   */
  public function __construct(
    CoreImageFactory $image_factory,
    LoggerChannelFactoryInterface $logger,
    TranslationInterface $string_translation,
    ConfigFactoryInterface $config_factory,
    EntityTypeManagerInterface $entity_type_manager,
    FileSystemInterface $file_system
  ) {
    $this->imageFactory = $image_factory;
    $this->logger = $logger;
    $this->setStringTranslation($string_translation);
    $this->configFactory = $config_factory;
    $this->entityTypeManager = $entity_type_manager;
    $this->fileSystem = $file_system;
    $this->quality = $config_factory->get('wpf.settings')->get('quality') ?? 75;
  }

  /**
   * {@inheritDoc}
   */
  public function createImageCopy(string $uri, ImageInterface $webp_image, $quality = NULL): bool {
    $this->uri = $uri;
    $webp_uri = $webp_image->getSource();
    $this->mimeType = 'image/jpeg';
    if ($quality !== NULL) {
      $this->quality = $quality;
    }

    $this->toolkitId = $this->configFactory->get('system.image')->get('toolkit');
    // Fall back to GD if the installed imagick does not exists.
    if ($this->toolkitId === 'imagick' && !extension_loaded('imagick')) {
      $this->toolkitId = 'gd';
    }

    // Get source image.
    $this->sourceImage = $this->imageFactory->get($this->uri, $this->toolkitId);
    // Get toolkit.
    $this->toolkit = $this->sourceImage->getToolkit();
    // Set destination uri.
    $this->setDestinationUri($webp_uri);

    if ($this->sourceImage === NULL ||
      $this->toolkit === NULL ||
      $this->destinationUri === NULL) {
      // Missing data to generate image.
      return FALSE;
    }

    if ($this->toolkitId === 'imagemagick' || $this->toolkitId === 'imagick') {
      // Using imagemagick to generate fallback jpg image.
      $image = $this->createImageMagickImage($webp_uri);
      return $image !== FALSE;
    }

    // Using GD to generate fallback jpg image.
    if ($this->toolkit instanceof GDToolkit) {
      /** @var \GdImage $gd_source_image */
      $gd_source_image = $this->toolkit->getResource();
      if ($gd_source_image !== NULL && $this->destinationUri !== NULL) {
        $success = FALSE;
        if ($image = imagecreatefromwebp($webp_uri)) {
          $success = imagejpeg($image, $this->destinationUri, $this->quality);
        }
        if ($success === FALSE) {
          $error = $this->t('Could not generate fallback image.');
          $this->logger->get('wpf')->error($error);
        }
        return $success;
      }
    }

    // Can't generate a GD resource from the source image, fail safely.
    $error = $this->t('Could not generate image resource from URI @uri.',
      ['@uri' => $uri]);
    $this->logger->get('wpf')->error($error);

    return FALSE;
  }

  /**
   * {@inheritDoc}
   */
  public function setDestinationUri(string $uri = NULL): void {
    if ($uri === NULL) {
      return;
    }
    $parts = explode('.', $uri);
    $last = array_key_last($parts);
    if ($last !== NULL && isset($parts[$last]) && $parts[$last] === 'webp') {
      $parts[$last] = 'jpg';
    }

    $this->destinationUri = implode('.', $parts);
  }

  /**
   * {@inheritDoc}
   */
  public function getSourceImage(): ImageInterface {
    return $this->sourceImage;
  }

  /**
   * {@inheritDoc}
   */
  public function getDestinationUri(): string {
    return $this->destinationUri;
  }

  /**
   * {@inheritDoc}
   */
  public function getMimeType(): string {
    return $this->mimeType;
  }

  /**
   * Generate image with imagemagick toolkit.
   *
   * @param string $uri
   *   The derivative image uri.
   *
   * @return false|string
   *   The destunation uri or false.
   */
  protected function createImageMagickImage(string $uri): bool|string {
    $imageMagickImg = $this->imageFactory->get($uri, $this->toolkit->getPluginId());
    // Convert webp image to jpg.
    $imageMagickImg->apply('convert',
      ['extension' => 'jpg', 'quality' => $this->quality]);

    $this->setDestinationUri($uri);
    $image = $this->destinationUri;
    if ($imageMagickImg->save($image)) {
      $msg = $this->t(
        'Generated WebP image with Image Magick. Quality: @quality Destination: @destination',
        [
          '@quality' => $this->quality,
          '@destination' => $image,
        ]
      );
      $this->logger->get('wpf')->info($msg);
    }
    else {
      $image = FALSE;
      $error = $this->t('Imagemagick issue: Could not generate WebP image.');
      $this->logger->get('wpf')->error($error);
    }

    return $image;
  }

  /**
   * @inheritDoc
   */
  public function fileDelete(EntityInterface $file): void {
    if (!$file instanceof FileInterface) {
      return;
    }
    $uri = $file->getFileUri();
    $styles = $this->entityTypeManager->getStorage('image_style')
      ->loadMultiple();
    /** @var \Drupal\image\Entity\Entity $style */
    foreach ($styles as $style) {
      $style_uri = $style->buildUri($uri);
      $copy_uri = _wpf_get_jpg($style_uri);
      if (file_exists($copy_uri)) {
        $this->fileSystem->delete($copy_uri);
      }
    }
  }

}
